/*
 * ASN.1 Teletexer, a TeletexString to Unicode converter.
 * (C) 2011 Sean Leonard of SeanTek(R) <dev+asn1teletexer@seantek.com>.
 * For licensing information, consult the LICENSE.txt or the author.
 *
 * version 0.7, 2011-07-27
 *
 * The following platforms are supported innately, in order:
 *  WIN32 (Windows)
 *  Darwin (Mac OS X/Carbon/Unicode Converter) (requires -framework CoreServices with gcc)
 *  iconv (libiconv, with HAVE_ICONV preprocessor directive) (requires -liconv or equivalent with gcc)
 *
 * If the platform has not been coded for, the conversion routines for JIS X 0208 and
 * GB 2312-80 will not be able to return the right results. Instead, the implementor can provide
 * alternative functions by adding the following preprocessor defines at the command-line:
 *  EXTERN_JISX0208_1997 nameofjpfunc
 *  EXTERN_GB2312_80_GB6345_1_86 nameofcnfunc
 * These defines need to point to valid function symbols at link time, with the C (not C++) signature:
 *  uint16_t dbcsfunc(const uint8_t b[2])
 * It is not a compiler error to compile without valid DBCS functions; however,
 * it will assert at runtime (if assert is turned on), or will return U+FFFD for all values
 * (if assert is turned off).
 * The functions must comply with outputting all results exactly as specified in
 *  tests/87-official.txt (JIS X 0208)
 *  tests/58-official.txt (GB 2312)
 * respectively.
 */

#include "asn1teletexer.h"
#include <assert.h>


typedef uint16_t (*sbcs_cb)(uint8_t b);
typedef uint16_t (*dbcs_cb)(const uint8_t b[2]);

static uint16_t C0_IR_1(uint8_t b) {
	assert(b <= 0x1F);
	return (uint16_t)b;
}

static uint16_t Teletex_C0_IR_106(uint8_t b) {
	assert(b <= 0x1F);
	assert((b & 0x7F) != 0x19 && (b & 0x7F) != 0x1D);
	return (uint16_t)b; /* same as C0_IR_1 */
}

static uint16_t Teletex_C1_IR_107_IR_77_IR_105(uint8_t b) {
	assert(b >= 0x80 && b <= 0x9F);
	return (uint16_t)b;
}

static uint16_t Teletex_Primary_IR_102_ASCII_IR_6(uint8_t b) {
	if (b == 0xA0 || b == 0xFF) return 0xFFFD; /* 94-character set */
	return b & 0x7F;
}

static uint16_t ISO6937_IR_156(uint8_t b);

/* This mapping relies on the union of IR 103 and IR 156,
   but preserves the 94-character set rule (if the byte is A0 or FF, it is bad). */
static uint16_t Teletex_Supplementary_IR_103(uint8_t b) {
	if (b == 0xA0 || b == 0xFF) return 0xFFFD; /* 94-character set */
	return ISO6937_IR_156(b);
}

/* This mapping is the union of IR 103 and IR 156 (ISO 6937),
   but does not handle combining diacritics. */
static uint16_t ISO6937_IR_156(uint8_t b) {
	uint8_t c = b & 0x7F; /* low char */
	assert((b & 0x7F) >= 0x20);
	assert((b & 0x70) != 0x40); /* don't handle combining diacritics */
	assert(c >= 0x20);

	if (c >= 0x60) {
		uint8_t i = c - 0x60;
		static const uint16_t hightable[] =
		{0x2126, 0xC6, 0x0110, 0xAA, 0x0126, 0xFFFD /* 0x0130 */, 0x0132, 0x013F,
		 0x0141, 0xD8, 0x0152, 0xBA, 0xDE, 0x0166, 0x014A, 0x0149,
		 0x0138, 0xE6, 0x0111, 0xF0, 0x0127, 0x0131, 0x0133, 0x0140,
		 0x0142, 0xF8, 0x0153, 0xDF, 0xFE, 0x0167, 0x014B, 0xAD};
		return hightable[i];
	} else if ((c & 0x70) == 0x50) {
		uint8_t i = c - 0x50;
		static const uint16_t table5[] = {0x2015, 0xB9, 0xAE, 0xA9, 0x2122, 0x266A, 0xAC, 0xA6};
		if (i < 8) return table5[i];
		else if (i >= 12) return i + 0x215B - 12; /* fractions */
		else return 0xFFFD;
	}
	if (c == 0x24) return 0x24;
	if (c == 0x34) return 0xD7;
	if (c == 0x26) return 0x23;
	if ((c & 0x7C) == 0x2C) return c + 0x2190 - 0x2C; /* arrows */
	else {
		uint8_t low4 = c & 0x0F;
		if (low4 >= 8 && low4 <= 10) {
			static const uint16_t column2[] = {0xA4, 0x2018, 0x201C};
			static const uint16_t column3[] = {0xF7, 0x2019, 0x201D};
			low4 -= 8;
			assert(c >= 0x28 && c <= 0x3A);
			return (c & 0x10) ? column3[low4] : column2[low4];
		}
		/* everybody else is the identity transform (with 8-bit set) */
		return c | 0x80;
	}
}

/* GB-Roman, aka GB 1988-80/GB 1988-89. */
static uint16_t GB_Roman_IR_57(uint8_t b) {
	if (b == 0xA0 || b == 0xFF) return 0xFFFD; /* 94-character set */
	else {
		uint8_t c = b & 0x7F;
		if (c == 0x24) return 0xA5;
		if (c == 0x7E) return 0x203E;
		return c;
	}
}

/* This mapping cross-references IR 126 and ISO 8859-7. Of note:
   0x24 = U+201C Euro
   0x25 = U+20AF Drachma
   0x2A = U+037A GREEK YPOGEGRAMMENI
   0x37 = U+00B7 MIDDLE DOT (sorry--after much review, everybody else converts
   to this and not U+0387 GREEK ANO TELIA so that is life)
   0x52 = U+FFFD (NOT U+03AA)
   Note that Windows-1253 is similar, but not identical, because of the currency symbols */
static uint16_t RH_Latin_Greek_IR_126(uint8_t b) {
	uint8_t c = b & 0x7F;
	assert(c >= 0x20);
	if (c >= 0x34) {
		if (c == 0x37) return 0xB7;
		if (c == 0x3B) return 0xBB;
		if (c == 0x3D) return 0xBD;
		if (c == 0x52 || c == 0x7F) return 0xFFFD;
		return c + 0x350;
	} else {
		static const uint16_t first6[] = {0xA0, 0x2018, 0x2019, 0xA3, 0x20AC, 0x20AF};
		if (c >= 0x30) return c | 0x80;
		if (c == 0x2F) return 0x2015;
		if (c == 0x2E) return 0xFFFD;
		if (c >= 0x26) {
			if (c == 0x2A) return 0x37A;
			else return c | 0x80;
		}
		if (c >= 0x20) return first6[c - 0x20];
		assert(0);
		return 0xFFFD; /* just in case */
	}
}

/* IR 144 Cyrillic, corresponds EXACTLY to ISO 8859-5. */
static uint16_t Cyrillic_IR_144(uint8_t b) {
	assert((b & 0x30) != 0);
	b |= 0x80;
	if (b == 0xA0 || b == 0xAD) return b;
	if (b == 0xF0) return 0x2116;
	if (b == 0xFD) return 0xA7;
	return b + 0x360;
}

/* IR 150 Greek, kind of a mix of ASCII ranges.
   Be aware that 0x52 is interpreted as CAPITAL GREEK LETTER SIGMA according to the note 2,
	 and 7E is MACRON (U+00AF). */
static uint16_t Greek_IR_150(uint8_t b) {
	assert((b & 0x30) != 0);
	if (b == 0x20 || b == 0x7F) return b; /* 94-character set */
	if (b == 0xA0 || b == 0xFF) return 0xFFFD;
	b &= 0x7F; /* constrict to 7-bit range */
	if (b == 0x60 || b == 0x5A || b == 0x7A) return 0xFFFD;
	if (b == 0x52) return 0x3A3;
	if (b == 0x7E) return 0xAF;
	if (b == 0xA4) return 0xA4;
	else {
		uint8_t b4 = b & 0x5F;
		if (b4 >= 0x41 && b4 <= 0x59) return b + 0x350;
		assert(b >= 0x21 && b <= 0x40 || b >= 0x5B && b <= 0x5F || b >= 0x7B && b <= 0x7D);
		return b;
	}
}

/* IR 153 Basic Cyrillic, from GOST-R. */
static uint16_t Basic_Cyrillic_IR_153(uint8_t b) {
	assert((b & 0x30) != 0);
	b |= 0x80;
	if (b >= 0xB0 && b <= 0xEF) return b + 0x360;
	if (b == 0xA0 || b == 0xAD) return b;
	if (b == 0xA1) return 0x401;
	if (b == 0xF1) return 0x451;
	return 0xFFFD;
}

/* IR 164 CITT Hebrew Supplementary Set, 96-character (no provision for SPACE, DELETE). */
static uint16_t Hebrew_IR_164(uint8_t b) {
	uint8_t c = b & 0x7F;
	assert((b & 0x30) != 0);
	if (c >= 0x60 && c <= 0x7A) return c + 0x05D0 - 0x60;
	return 0xFFFD;
}

static uint16_t Latin1_IR_100(uint8_t b) {
	return b | 0x80;
}

/* double-byte character sets */

#if defined(EXTERN_JISX0208_1997) && defined(EXTERN_GB2312_80_GB6345_1_86)

uint16_t EXTERN_JISX0208_1997(const uint8_t b[2]);

static uint16_t JISX0208_1997(const uint8_t b[2]) {
	return EXTERN_JISX0208_1997(b);
}

uint16_t EXTERN_GB2312_80_GB6345_1_86(const uint8_t b[2]);

static uint16_t GB2312_80_GB6345_1_86(const uint8_t b[2]) {
	return EXTERN_GB2312_80_GB6345_1_86(b);
}

#elif defined(WIN32)

#ifndef _WIN32_WINNT
#define _WIN32_WINNT 0x0500
#endif
#ifndef WINVER
#define WINVER 0x0500
#endif

// #include <winnt.h>
// #define NT_INCLUDED
// #include <windef.h>

#ifndef MB_ERR_INVALID_CHARS
#define MB_ERR_INVALID_CHARS      0x00000008
#endif

#if (defined(_M_IX86) || defined(_M_IA64) || defined(_M_AMD64)) && !defined(MIDL_PASS)
#define DECLSPEC_IMPORT __declspec(dllimport)
#else
#define DECLSPEC_IMPORT
#endif

#ifdef _MAC
#define WINAPI      CDECL
#elif (_MSC_VER >= 800) || defined(_STDCALL_SUPPORTED)
#define WINAPI      __stdcall
#else
#define WINAPI
#endif

typedef unsigned long       DWORD;

#ifndef _MAC
typedef wchar_t WCHAR;
#else
typedef unsigned short WCHAR;
#endif

#pragma comment(lib, "kernel32.lib")

DECLSPEC_IMPORT
int
WINAPI
MultiByteToWideChar(unsigned int CodePage, DWORD dwFlags, const char *lpMultiByteStr,
	int cbMultiByte, WCHAR *lpWideCharStr, int cchWideChar);

/* JIS X 0208-1983 and 1990. See notes on stub function. */
static uint16_t JISX0208_1997(const uint8_t b[2]) {
	/* char bytes[5] = {0x1B, 0x24, 0x42, (char)(b[0] & 0x7F), (char)(b[1] & 0x7F)}; */
	uint8_t bytes[2] = {b[0] | 0x80, b[1] | 0x80};
	WCHAR outchars[2] = {0};
	int mb2wret;
	/* use EUC-JP because we don't need escape characters.
	 * 20932 actually uses code page 932's repetoire. However, this is not just JIS X 0208.
	 * Row 13 and rows 89-92 are assigned to characters in NEC Row 13 and IBM Selected Rows 89-92.
	 * These are dropped. Compared to the Windows status quo:
	 * 01-29 U+2014
	 * 01-81 U+FFE0
	 * 01-82 U+FFE1
	 * 02-44 U+FFE2
	 */
	if ((bytes[0] < 0xB0 && bytes[0] > 0xA8) || bytes[0] > 0xF4)
		return 0xFFFD;
	if (bytes[0] == 0xA1) {
		if (bytes[1] == 0xBD) return 0x2014;
		else if (bytes[1] == 0xF1) return 0xFFE0;
		else if (bytes[1] == 0xF2) return 0xFFE1;
	} else if (bytes[0] == 0xA2 && bytes[1] == 0xCC) return 0xFFE2; /* not sign */

	mb2wret = MultiByteToWideChar(20932, MB_ERR_INVALID_CHARS, (char*)bytes, 2, outchars, 2);

	if (mb2wret == 0)
		return 0xFFFD;

	assert(mb2wret == 1);
	/* only 01-06 should map to U+30FB KATAKANA MIDDLE DOT */
	assert(outchars[0] != 0x30FB || bytes[0] == 0xA1 && bytes[1] == 0xA6);
	if (!(outchars[0] != 0x30FB || bytes[0] == 0xA1 && bytes[1] == 0xA6))
		return 0xFFFD;
	return outchars[0];
}

/* row-cell (QuWei) to byte value */
#define QW2B(qw) (qw + 0xA0)

/* GB 2312-80 with GB 6345.1-86 additions. Actually follows MS CP 936 to a large extent. */
static uint16_t GB2312_80_GB6345_1_86(const uint8_t b[2]) {
	uint8_t bytes[2] = {b[0] | 0x80, b[1] | 0x80};
	WCHAR outchars[2] = {0};
	int mb2wret;

	if (bytes[0] == QW2B(11)) { /* half-width pinyin characters (same as row 8, for our purposes) */
		if (bytes[1] > QW2B(32)) return 0xFFFD;
		else bytes[0] -= 3;
	} else if (bytes[0] == QW2B(10)) { /* ISO646-CN */
		if (bytes[1] == QW2B(4)) return 0xA5; /* Yen Sign */
		else if (bytes[1] == QW2B(94)) return 0x203E; /* Overline */
		else return (uint16_t)(bytes[1] - 0x80);
	} else if (bytes[0] >= QW2B(88)) return 0xFFFD; /* 88-94 are unassigned */
	else if (bytes[0] == QW2B(2) && bytes[1] < 0xB1) return 0xFFFD; /* don't emit lowercase roman numerals */

	if (bytes[0] == QW2B(8)) { /* "full-width" pinyin and bopomofo */
		if (bytes[1] == 0xBC) return 0x1E3F;
		if (bytes[1] == 0xBF) return 0x01F9;
	}
	else if (bytes[0] <= QW2B(15) && bytes[0] >= QW2B(12)) return 0xFFFD;

	mb2wret = MultiByteToWideChar(936, MB_ERR_INVALID_CHARS, (char*)bytes, 2, outchars, 2);

	if (mb2wret == 0) return 0xFFFD;

	assert(mb2wret == 1);
	if (outchars[0] >= 0xE000 && outchars[0] <= 0xF8FF) return 0xFFFD; /* no PUA allowed */

	return outchars[0];
}

#undef QW2B

#elif defined(__APPLE__)

#include <CoreServices/CoreServices.h>

static uint16_t JISX0208_1997(const uint8_t b[2]) {
	OSStatus s;
	TextToUnicodeInfo tuInfo;
	UnicodeMapping uMapping;
	uint8_t bytes[2] = {b[0] | 0x80, b[1] | 0x80};
	UniChar outchars[2] = {0};
	ByteCount sourceRead;
	ByteCount unicodeLen;

	if ((bytes[0] < 0xB0 && bytes[0] > 0xA8) || bytes[0] > 0xF4) return 0xFFFD;

	uMapping.unicodeEncoding = CreateTextEncoding(kTextEncodingUnicodeV3_0, kUnicodeNoSubset, kUnicode16BitFormat);
	uMapping.otherEncoding = CreateTextEncoding(kTextEncodingEUC_JP, kTextEncodingDefaultVariant, kTextEncodingDefaultFormat);
	uMapping.mappingVersion = kUnicodeUseLatestMapping;
	
	s = CreateTextToUnicodeInfo(&uMapping, &tuInfo);
	assert(s == 0);
	if (s != 0) return 0xFFFD;
	
	s = ConvertFromTextToUnicode(tuInfo, 2, bytes, 0, 0, NULL, NULL, NULL, 2,
		&sourceRead, &unicodeLen, outchars);
	if (s != 0) {
		DisposeTextToUnicodeInfo(&tuInfo);
		return 0xFFFD;
	}
	assert(unicodeLen / sizeof(UniChar) == 1);
	s = DisposeTextToUnicodeInfo(&tuInfo);
	assert(s == 0);

	return outchars[0];
}

/* row-cell (QuWei) to byte value */
#define QW2B(qw) (qw + 0xA0)

static uint16_t GB2312_80_GB6345_1_86(const uint8_t b[2]) {
	OSStatus s;
	TextToUnicodeInfo tuInfo;
	UnicodeMapping uMapping;
	uint8_t bytes[2] = {b[0] | 0x80, b[1] | 0x80};
	UniChar outchars[2] = {0};
	ByteCount sourceRead;
	ByteCount unicodeLen;
	
	if (bytes[0] == QW2B(11)) { /* half-width pinyin characters (same as row 8, for our purposes) */
		if (bytes[1] > QW2B(32)) return 0xFFFD;
		else bytes[0] -= 3;
	} else if (bytes[0] == QW2B(10)) { /* ISO646-CN */
		if (bytes[1] == QW2B(4)) return 0xA5; /* Yen Sign */
		else if (bytes[1] == QW2B(94)) return 0x203E; /* Overline */
		else return (uint16_t)(bytes[1] - 0x80);
	} else if (bytes[0] >= QW2B(88)) return 0xFFFD; /* 88-94 are unassigned */
	else if (bytes[0] == QW2B(3)) {
		if (bytes[1] == QW2B(4)) return 0xFFE5; /* Fullwidth Yen Sign */
		else if (bytes[1] == QW2B(94)) return 0xFFE3; /* Fullwidth Macron (Overline) */
	}

	uMapping.unicodeEncoding = CreateTextEncoding(kTextEncodingUnicodeV3_0, kUnicodeNoSubset, kUnicode16BitFormat);
	uMapping.otherEncoding = CreateTextEncoding(kTextEncodingEUC_CN, kEUC_CN_BasicVariant, kTextEncodingDefaultFormat);
	uMapping.mappingVersion = kUnicodeUseLatestMapping;

	s = CreateTextToUnicodeInfo(&uMapping, &tuInfo);
	assert(s == 0);
	if (s != 0) return 0xFFFD;
	
	s = ConvertFromTextToUnicode(tuInfo, 2, bytes, 0, 0, NULL, NULL, NULL, 2,
		&sourceRead, &unicodeLen, outchars);
	if (s != 0) {
		DisposeTextToUnicodeInfo(&tuInfo);
		return 0xFFFD;
	}
	assert(unicodeLen / sizeof(UniChar) == 1);
	s = DisposeTextToUnicodeInfo(&tuInfo);
	assert(s == 0);

	return outchars[0];
}

#undef QW2B

#elif defined(HAVE_ICONV)

#include <iconv.h>

static uint16_t JISX0208_1997(const uint8_t b[2]) {
	uint8_t bytes[5] = {0x1B, '$', 'B', b[0] & 0x7F, b[1] & 0x7F};
	
	uint16_t outchars[2] = {0};
	char *inbuf = (char*)bytes, *outbuf = (char*)outchars;
	size_t inbytesleft = 5, outbytesleft = 2 * sizeof(uint16_t), ret;
	iconv_t iv;
	
	if (bytes[3] == 0x21) {
		if (bytes[4] == 0x3D) return 0x2014;
		else if (bytes[4] == 0x71) return 0xFFE0;
		else if (bytes[4] == 0x72) return 0xFFE1;
	} else if (bytes[3] == 0x22 && bytes[4] == 0x4C) return 0xFFE2; /* not sign */
	
	iv = iconv_open("UCS-2-INTERNAL", "ISO-2022-JP");
	assert(iv != (iconv_t)-1);
	if (iv == (iconv_t)-1) return 0xFFFD;
	
	ret = iconv(iv, &inbuf, &inbytesleft, &outbuf, &outbytesleft);
	iconv_close(iv);
	
	if (ret == (size_t)-1 || outbytesleft != 1 * sizeof(uint16_t)) return 0xFFFD;
	else return outchars[0];
}

static uint16_t GB2312_80_GB6345_1_86(const uint8_t b[2]) {
	uint8_t bytes[7] = {0x1B, '$', ')', 'A', 0x0E, b[0] & 0x7F, b[1] & 0x7F};

	uint16_t outchars[2] = {0};
	char *inbuf = (char*)bytes, *outbuf = (char*)outchars;
	size_t inbytesleft = sizeof(bytes), outbytesleft = 2 * sizeof(uint16_t), ret;
	iconv_t iv;
	
	if (bytes[5] == 0x2B) {
		if (bytes[6] > 32 + 0x21) return 0xFFFD;
		else bytes[5] -= 3;
	} else if (bytes[5] == 0x2A) {
		if (bytes[6] == 0x24) return 0xA5;
		else if (bytes[6] == 0x7E) return 0x203E;
		else return (uint16_t)bytes[6];
	}
	
	if (bytes[5] == 0x21) {
		if (bytes[6] == 0x24) return 0xB7;
		else if (bytes[6] == 0x2A) return 0x2014;
	} else if (bytes[5] == 0x26) {
		static const uint16_t brackets[] = {0xFE35, 0xFE36, 0xFE39, 0xFE3A, 0xFE3F, 0xFE40, 0xFE3D, 0xFE3E, 0xFE41,
		 0xFE42, 0xFE43, 0xFE44, 0xFFFD, 0xFFFD, 0xFE3B, 0xFE3C, 0xFE37, 0xFE38, 0xFE31, 0xFFFD, 0xFE33, 0xFE34};
		if (bytes[6] > 0x5F && bytes[6] < 0x76) return brackets[bytes[6] - 0x60];
	} else if (bytes[5] == 0x28) {
		static const uint16_t sixchars[6] = {0x251, 0x1E3F, 0x144, 0x148, 0x1F9, 0x261};
		if (bytes[6] >= 0x3B && bytes[6] <= 0x40) return sixchars[bytes[6] - 0x3B];
	}

	iv = iconv_open("UCS-2-INTERNAL", "ISO-2022-CN");
	assert(iv != (iconv_t)-1);
	if (iv == (iconv_t)-1) return 0xFFFD;

	ret = iconv(iv, &inbuf, &inbytesleft, &outbuf, &outbytesleft);
	iconv_close(iv);

	if (ret == (size_t)-1 || outbytesleft != 1 * sizeof(uint16_t)) return 0xFFFD;
	else return outchars[0];
}

/* for subsequent use below */
#define __USED_ICONV_IN_TELETEXER

#else

#if defined(_MSC_VER)
/* MS KB155196 */
#define __STR2__(x) #x
#define __STR1__(x) __STR2__(x)
/* in VS 2010 this gets converted to "asn1teletexer.c(#): warning CEACNV: Japanese (JISX0208_1997)..."
   (with different spacing) */
#define __LOC__ __FILE__ "("__STR1__(__LINE__)") : Warning CEACNV : "
#pragma message(__LOC__"Japanese (JISX0208_1997) and Chinese (GB2312_80_GB6345_1_86) are not implemented on this platform.")
#elif defined(__GNUC__)
#warning Japanese (JISX0208_1997) and Chinese (GB2312_80_GB6345_1_86) are not implemented on this platform.
#endif

/* JIS X 0208-1983 and 1990 (effectively JIS X 0208:1997).
   The only differences are that 1990 has kuten bango 84-05 and 84-06.
	 This should match the repetoire of the 1997 spec exactly, with no codes assigned
	 to rows 13 or 89-92 (unlike MS CP 932).
 */
static uint16_t JISX0208_1997(const uint8_t b[2]) {
	assert(0);
	return 0xFFFD;
}

/* GB2312-80, including GB6345.1-86 corrections and additions. See Ken Lunde,
   CJKV Information Processing, Second Edition (2009) pgs. 94-97.
	 The behavior of this call is that it must return the results as of GB 6345.1-86,
	 assigning all invalid bit combinations to U+FFFD. Row 2 columns 1-16 must be invalid
	 (no Roman i-xii). Row 8 must include all 32 pinyin characters. Rows 10-11 must map properly.
 */
static uint16_t GB2312_80_GB6345_1_86(const uint8_t b[2]) {
	assert(0);
	return 0xFFFD;
}

#endif

#define JISX0208_1983_IR_87 JISX0208_1997
#define JISX0208_1990_IR_168 JISX0208_1997

#define GB2312_80_IR_58 GB2312_80_GB6345_1_86

#if defined(HAVE_ICONV) && defined(__USED_ICONV_IN_TELETEXER)

/* libiconv version, which should only be around if the penultimate if block above is on */
static uint16_t Chinese_IR_165(const uint8_t b[2]) {
	uint8_t r = (b[0] & 0x7F) - 0x20; /* row */
	if (r < 12 || r >= 16 && r < 90) return GB2312_80_IR_58(b);
	else {
		uint8_t bytes[7] = {0x1B, '$', ')', 'E', 0x0E, b[0] & 0x7F, b[1] & 0x7F};

		uint16_t outchars[2] = {0};
		char *inbuf = (char*)bytes, *outbuf = (char*)outchars;
		size_t inbytesleft = sizeof(bytes), outbytesleft = 2 * sizeof(uint16_t), ret;
		iconv_t iv;
		
		iv = iconv_open("UCS-2-INTERNAL", "ISO-2022-CN-EXT");
		assert(iv != (iconv_t)-1);
		if (iv == (iconv_t)-1) return 0xFFFD;

		ret = iconv(iv, &inbuf, &inbytesleft, &outbuf, &outbytesleft);
		iconv_close(iv);
		if (ret == (size_t)-1 || outbytesleft != 1 * sizeof(uint16_t)) return 0xFFFD;
		else return outchars[0];
	}
}

#else

/* Chinese ISO IR 165. Equivalent to CCITT T.101 Table C.16. Includes GB 8565.2-88 and unique #165 additions.
   This set requires tables because virtually no major platform supports the full set.
	 It fills in rows 12-15 and 90-94 on its own (but not row 6), otherwise deferring to the platform for #58.
	 See Ken Lunde, CJKV Information Processing, Second Edition (2009) pgs. 98-99.
 */
static uint16_t Chinese_IR_165(const uint8_t b[2]) {
	uint8_t r = (b[0] & 0x7F) - 0x20; /* row */
	uint8_t c = (b[1] & 0x7F) - 0x20; /* column */

	assert(r >= 1 && r <= 94 && c >= 1 && c <= 94);

	/* Row 6 "22 background (shading) characters" *60-81) are intentionally ignored. The majority are not representable in Unicode
	   and apparently no other platform supports them anyway. If a sender is sending these characters, it is probably an attack. */
	if (r < 12) return GB2312_80_IR_58(b);
	else if (r < 15) { /* GB 8565.2-88 and ISO-IR-165 additions */
		static const uint16_t row12map[3][94] = {
{0x53BE, 0x4EB8, 0x4F3E, 0x501E, 0x50C7, 0x9118, 0x6C98, 0x6CDC, 0x6CC3, 0x6E5D, 0x6EA6, 0x6EEB, 0x6FA5, 0x6165, 0x5EA4, 0x9618, 0x5848, 0x8453, 0x7CF5,
 0x5F07, 0x6294, 0x647D, 0x725A, 0x5574, 0x55A4, 0x5640, 0x5684, 0x5D1F, 0x72C9, 0x998C, 0x59DE, 0x59FD, 0x5A5E, 0x7EBB, 0x7EE4, 0x7EF9, 0x9A99, 0x71CF,
 0x6245, 0x624A, 0x797C, 0x739A, 0x742B, 0x7488, 0x74AA, 0x74D8, 0x6767, 0x6AB5, 0x71CA, 0x6BA3, 0x8F80, 0x8F92, 0x8D5F, 0x9B36, 0x72A8, 0x87A3, 0x8152,
 0x6B38, 0x98D0, 0x8897, 0x88AF, 0x8955, 0x770A, 0x94DA, 0x955A, 0x9560, 0x9E24, 0x9E40, 0x9E50, 0x9E5D, 0x9E60, 0x870E, 0x7B5C, 0x7FD9, 0x7FEF, 0x7E44,
 0x8E45, 0x8E36, 0x8E62, 0x8E5C, 0x9778, 0x9B46, 0x9F2B, 0x9F41, 0x7526, 0x4E26, 0x8BAC, 0x8129, 0x5091, 0x50CD, 0x52B9, 0x89D4, 0x5557, 0x94C7},
{0x9496, 0x9498, 0x94CF, 0x94D3, 0x94D4, 0x94E6, 0x9533, 0x951C, 0x9520, 0x9527, 0x953D, 0x9543, 0x956E, 0x9574, 0x9C80, 0x9C84, 0x9C8A, 0x9C93, 0x9C96,
 0x9C97, 0x9C98, 0x9C99, 0x9CBF, 0x9CC0, 0x9CC1, 0x9CD2, 0x9CDB, 0x9CE0, 0x9CE3, 0x9770, 0x977A, 0x97A1, 0x97AE, 0x97A8, 0x9964, 0x9966, 0x9978, 0x9979,
 0x997B, 0x997E, 0x9982, 0x9983, 0x998E, 0x9B10, 0x9B18, 0x65A2, 0x9E80, 0x911C, 0x9E91, 0x9F12, 0x52F3, 0x6C96, 0x6D44, 0x6E1B, 0x6E67, 0x6F82, 0x6FEC,
 0x60AE, 0x5EC8, 0x8FFA, 0x577F, 0x5586, 0x849E, 0x8460, 0x5C05, 0x5E0B, 0x5D11, 0x5D19, 0x5DD6, 0x59B3, 0x5AAE, 0x9A94, 0x658F, 0x709E, 0x7551, 0x71FF,
 0x691D, 0x6A11, 0x68BF, 0x6607, 0x668E, 0x6673, 0x6C25, 0x7652, 0x778B, 0x76EA, 0x9895, 0x8780, 0x882D, 0x7B87, 0x7C50, 0x8EAD, 0x9575, 0x65C2},
{0x5390, 0x79B8, 0x4F15, 0x4F21, 0x4F3B, 0x4FA2, 0x50A4, 0x5092, 0x530A, 0x51C3, 0x51A8, 0x8D20, 0x5787, 0x579A, 0x5795, 0x57EB, 0x585D, 0x585A, 0x5871,
 0x5895, 0x5C30, 0x5F0C, 0x5F0D, 0x5F0E, 0x5C72, 0x5CC7, 0x5FAC, 0x5F68, 0x5F5F, 0x5A12, 0x5A65, 0x5A84, 0x5AC4, 0x7394, 0x73EA, 0x73EE, 0x7437, 0x7415,
 0x7454, 0x6799, 0x686C, 0x68F8, 0x69FE, 0x72E2, 0x6667, 0x8D52, 0x89C3, 0x89CD, 0x6427, 0x6477, 0x6C1D, 0x813F, 0x6B54, 0x98D6, 0x707A, 0x70F1, 0x7120,
 0x6153, 0x6C87, 0x6DAD, 0x6E81, 0x6EB5, 0x6F94, 0x6F9B, 0x793D, 0x794E, 0x7806, 0x7859, 0x7894, 0x78DC, 0x7903, 0x7A16, 0x7A5E, 0x75E0, 0x7ADC, 0x7676,
 0x9892, 0x7BF2, 0x7C30, 0x7C5D, 0x9C9D, 0x7CAC, 0x8278, 0x83D1, 0x84EA, 0x7FC0, 0x7F1E, 0x8E21, 0x8E53, 0x9754, 0x9F0C, 0x94FB, 0xFFFD, 0xFFFD}
};
		return row12map[r - 12][c - 1];
	} else if (r < 16) { /* more GB 8565.2-88 and ISO-IR-165 additions, but partially algorithmic */
		if (c <= 12) return c + 0x32C0 - 1; /* 12 months */
		if (c <= 43) return c + 0x33E0 - 13; /* 31 days */
		if (c <= 68) return c + 0x3358 - 44; /* 24 points */
		else {
			static const uint16_t row15endmap[26] = {0x3037 /* XX */, 0x90A8, 0x965E, 0x5842, 0x5803, 0x6C3E, 0x6D29, 0x6EE7,
			 0x8534, 0x84C6, 0x633C, 0x5D05, 0x7F10, 0x7EEC, 0x7287, 0x712E, 0x8218, 0x8216, 0x756C, 0x75F3, 0x9B25, 0x8980, 0x7CA6, 0x4E85, 0x5570, 0x91C6};
			return row15endmap[c - 69];
		}
	} else if (r < 90) return GB2312_80_IR_58(b);
	else { /* GB 8565.2-88 additions */
		/* no way around it: there are no consistent, algorithmic patterns greater than a few code points between this and Unicode, or this and GBK */
		static const uint16_t row90map[5][94] = {
{0x4E0F, 0x673F, 0x4E42, 0x752A, 0x592C, 0x9EE1, 0x8652, 0x531C, 0x5187, 0x518F, 0x50F0, 0x4F0B, 0x4F23, 0x4F03, 0x4F61, 0x4F7A, 0x4F6B, 0x4FEB, 0x4FF5,
 0x5034, 0x5022, 0x4FF6, 0x5072, 0x4EB6, 0x51AE, 0x5910, 0x6BDA, 0x522C, 0x5232, 0x4FB4, 0x5298, 0x52BB, 0x52BC, 0x52CD, 0x52DA, 0x52F7, 0x53C6, 0x53C7,
 0x5770, 0x576C, 0x57B1, 0x579F, 0x579E, 0x57BE, 0x57CC, 0x580E, 0x580C, 0x57F5, 0x5809, 0x583C, 0x5843, 0x5845, 0x5846, 0x583D, 0x5853, 0x5888, 0x5884,
 0x58F8, 0x56AD, 0x5940, 0x5953, 0x596D, 0x5C2A, 0x54A5, 0x551D, 0x5536, 0x556F, 0x554D, 0x569A, 0x569C, 0x56F7, 0x5710, 0x5719, 0x5E17, 0x5E21, 0x5E28,
 0x5E6A, 0x5C74, 0x5C7C, 0x5CA8, 0x5C9E, 0x5CC3, 0x5CD3, 0x5CE3, 0x5CE7, 0x5CFF, 0x5D04, 0x5D00, 0x5D1A, 0x5D0C, 0x5D4E, 0x5D5A, 0x5D85, 0x5D93},
{0x5D92, 0x5DC2, 0x5DC9, 0x8852, 0x5FAF, 0x5906, 0x65A8, 0x7241, 0x7242, 0x5EBC, 0x5ECB, 0x95EC, 0x95FF, 0x8A1A, 0x9607, 0x9613, 0x961B, 0x5BAC, 0x5BA7,
 0x5C5D, 0x5F22, 0x59EE, 0x5A7C, 0x5A96, 0x5A73, 0x5A9E, 0x5AAD, 0x5ADA, 0x5AEA, 0x5B1B, 0x5B56, 0x9A72, 0x9A83, 0x9A89, 0x9A8D, 0x9A8E, 0x9A95, 0x9AA6,
 0x7395, 0x7399, 0x73A0, 0x73B1, 0x73A5, 0x73A6, 0x73D6, 0x73F0, 0x73FD, 0x73E3, 0x7424, 0x740E, 0x7407, 0x73F6, 0x73FA, 0x7432, 0x742F, 0x7444, 0x7442,
 0x7471, 0x7478, 0x7462, 0x7486, 0x749F, 0x74A0, 0x7498, 0x74B2, 0x97E8, 0x6745, 0x679F, 0x677B, 0x67C8, 0x67EE, 0x684B, 0x68A0, 0x6812, 0x681F, 0x686A,
 0x68BC, 0x68FB, 0x686F, 0x68B1, 0x68C1, 0x68EB, 0x6913, 0x68D1, 0x6911, 0x68D3, 0x68EC, 0x692B, 0x68E8, 0x69BE, 0x6969, 0x6940, 0x696F, 0x695F},
{0x6962, 0x6935, 0x6959, 0x69BC, 0x69C5, 0x69DA, 0x69DC, 0x6A0B, 0x69E5, 0x6A66, 0x6A96, 0x6AB4, 0x72DD, 0x5CF1, 0x7314, 0x733A, 0x6B95, 0x5F67, 0x80FE,
 0x74FB, 0x7503, 0x655C, 0x6569, 0x6F26, 0x65F8, 0x65FB, 0x6609, 0x663D, 0x6662, 0x665E, 0x666C, 0x668D, 0x668B, 0x8D51, 0x8D57, 0x7263, 0x7277, 0x63B1,
 0x6261, 0x6260, 0x6283, 0x62E4, 0x62C3, 0x631C, 0x6326, 0x63AF, 0x63FE, 0x6422, 0x6412, 0x64ED, 0x6713, 0x6718, 0x8158, 0x81D1, 0x98CF, 0x98D4, 0x98D7,
 0x6996, 0x7098, 0x70DC, 0x70FA, 0x710C, 0x711C, 0x71CB, 0x721F, 0x70DD, 0x659D, 0x6246, 0x6017, 0x60C7, 0x60D3, 0x60B0, 0x60D9, 0x6114, 0x6C3F, 0x6C67,
 0x6C84, 0x6C9A, 0x6C6D, 0x6CA8, 0x6CC6, 0x6CB5, 0x6D49, 0x6D38, 0x6D11, 0x6D3A, 0x6D28, 0x6D50, 0x6D34, 0x6D55, 0x6D61, 0x6DA2, 0x6D65, 0x6D5B},
{0x6D64, 0x6DB4, 0x6E9A, 0x6E5C, 0x6E72, 0x6EA0, 0x6E87, 0x6E8E, 0x6EC9, 0x6EC3, 0x6F37, 0x6ED8, 0x6EEA, 0x6F56, 0x6F75, 0x6F5F, 0x6FB4, 0x6FBC, 0x7014,
 0x700D, 0x700C, 0x703C, 0x7943, 0x7947, 0x794A, 0x7950, 0x7972, 0x7998, 0x79A0, 0x79A4, 0x77FC, 0x77FB, 0x7822, 0x7820, 0x7841, 0x785A, 0x7875, 0x78B6,
 0x78E1, 0x7933, 0x8A5F, 0x76FB, 0x771B, 0x772C, 0x7786, 0x77AB, 0x77AD, 0x7564, 0x756F, 0x6983, 0x7F7D, 0x76DD, 0x76E6, 0x76EC, 0x7521, 0x79FE, 0x7A44,
 0x767F, 0x769E, 0x9E27, 0x9E2E, 0x9E30, 0x9E34, 0x9E4D, 0x9E52, 0x9E53, 0x9E54, 0x9E56, 0x9E59, 0x9E61, 0x9E62, 0x9E65, 0x9E6F, 0x9E74, 0x75A2, 0x7604,
 0x7608, 0x761D, 0x7AD1, 0x7A85, 0x7A8E, 0x7AA3, 0x7AB8, 0x7ABE, 0x77DE, 0x8030, 0x988B, 0x988E, 0x9899, 0x98A3, 0x8683, 0x8705, 0x8758, 0x87CF},
{0x87E2, 0x880B, 0x80D4, 0x7F4D, 0x7B4A, 0x7B4E, 0x7B7F, 0x7B93, 0x7BEF, 0x7C09, 0x7BF0, 0x7C15, 0x7C03, 0x7C20, 0x823A, 0x8886, 0x88AA, 0x88C0, 0x88C8,
 0x8926, 0x8976, 0x7F91, 0x8283, 0x82BC, 0x82A7, 0x8313, 0x82FE, 0x8300, 0x835D, 0x8345, 0x8344, 0x831D, 0x83A6, 0x8399, 0x83FE, 0x841A, 0x83FC, 0x8429,
 0x8439, 0x84A8, 0x84CF, 0x849F, 0x84C2, 0x84F7, 0x8570, 0x85B3, 0x85A2, 0x96D8, 0x85B8, 0x85E0, 0x7FDA, 0x7EAE, 0x7EB4, 0x7EBC, 0x7ED6, 0x7F0A, 0x5B43,
 0x8D6A, 0x5245, 0x8C68, 0x8C6E, 0x8C6D, 0x8E16, 0x8E26, 0x8E27, 0x8E50, 0x9098, 0x90A0, 0x90BD, 0x90C8, 0x90C3, 0x90DA, 0x90FF, 0x911A, 0x910C, 0x9120,
 0x9142, 0x8FB5, 0x90E4, 0x8C86, 0x89F1, 0x8BB1, 0x8BBB, 0x8BC7, 0x8BEA, 0x8C09, 0x8C1E, 0x9702, 0x68D0, 0x7306, 0x9F81, 0x9F82, 0x92C6, 0x9491}
};
		return row90map[r - 90][c - 1];
	}
}

#endif

static uint16_t To_T61_Diacritic(uint8_t b) {
	assert(b >= 0x41 && b <= 0x4F && b != 0x49 && b != 0x4C);
	if (b <= 0x45) return b + 0x2BF;
	else if (b <= 0x4A) return b + 0x2C0;
	else if (b == 0x4B) return 0x327;
	else if (b == 0x4D) return 0x30B;
	else if (b == 0x4E) return 0x328;
	else {
		assert(b == 0x4F);
		return 0x30C;
	}
}

static int Is_T61_Latin(uint16_t latin) {
	if (latin == 0x20) return 1;
	else {
		const uint8_t latinfold = latin & 0x5F;
		if (latinfold >= 0x41 && latinfold <= 0x5A) return 1;
		return 0;
	}
}

#define buf (*_buf)
#define wbuf (*_wbuf)
static int T61_Diacritical_Worms(const uint8_t **_buf, uint16_t **_wbuf,
 const uint8_t * const endbuf, const sbcs_cb G[4], const dbcs_cb GDB[4],
 const unsigned char GLR[2]) {
	/* can of worms regarding combining diacritics */
	uint8_t b = *buf, b2;
	sbcs_cb g2;
	uint8_t c = b & 0x7F; /* c is b with top bit cut off */
	if (c == 0x40 || c == 0x49) *wbuf = 0xFFFD; /* soft error */
	/* no C0 or C1, not even the pathological case of underline + PLD/PLU */
	else if (buf++ >= endbuf || ((((b2 = *buf) & 0x60) == 0) /* &&
		c != 0x4C && (b2 != 0x8B || b2 != 0x8C) */)) return 0;
	else if (c == 0x4C) { /* underline */
		unsigned char glr2;
		/* There is only one kind of C1 supported (unified #77, #105, #107) so we don't need to check.
				#103 says in note 4 "The non-spacing character UNDERLINE can be used in combination with any graphic
				character of the repetoire, including an accented letter or an umlaut, or SPACE."
				Unlike note 3, no reference is made to T.61.

				T.61:1988 says that PLD/PLU are POSSIBLE! Specifically, 4.1.3.1 e) says "Between the Non-spacing
				underline character and the character to which it applies, one or more control functions may occur,
				e.g. Partial line down (PLD), Partial line up (PLU) or a code extension control function."
				A code extension control function is only CSI, followed by some bytes. Other "control functions"
				theoretically are possible. However, we can finesse this away by noting that note 4 does not
				actually reference T.61, that T.61 itself is withdrawn, and that X.680:2008 and X.690:2009 omit
				all possible references to the T.61 specification.
				In any event, the combination _ PLD char means (3.3.2 CF16):
				 "PLD does not affect the vertical position of any underlining of subsequent
				 character(s) if the underlining is invoked (by SGR or Non-spacing underline)
				 prior to the PLD."
				See also CP03. The idea is that the underline is drawn at the old vertical position, and then
				the character is drawn at the PLD (lowered) position (without drawing a "lower" underline).
				This concept is literally unrepresentable in Unicode: U+0332 is a combining character
				that only applies to the prior character (thus, representing an underlined character), not to
				any graphic command (not to underline an area at a different vertical position than the character).
				Additionally, X.690 8.23.5.1 says "An escape sequence shall not be used unless it is one of those
				specified by one of the registration numbers used to define the character string type in X.680."
				Thus all escape sequences outside of the "norm" (used to designate the character set) are fair
				game for removal.
				Given this analysis, the best & most accurate interpretation is only to accept graphic chars
				after the underline.
					
		if (b2 == 0x8B || b2 == 0x8C) {
			*wbuf++ = (uint16_t)b2;
			if (buf++ >= endbuf) return 0;
			b2 = *buf;
		} */
		glr2 = GLR[b2 >> 7];
		g2 = G[glr2];
		/* pathological case would be underline plus diacritic in OTHER GL/GR.
			It's a near-impossible combination, EXCEPT for the case where GR=UNDERLINE,
			GL=other diacritic, and finally GL=SPACE. */
		if ((g2 == Teletex_Supplementary_IR_103 || g2 == ISO6937_IR_156) && (b2 & 0x70) == 0x40) {
			uint8_t c2 = b2 & 0x7F;
			uint16_t latin;
			/* Deal with diacritic plus graphic. The following if statement is horrifying
					yet very efficient. Compare with the diacritic+latin processing below. */
			if (c2 == 0x40 || c2 == 0x49 || c2 == 0x4C || buf++ >= endbuf ||
				((b2 = *buf) & 0x60) == 0 || (g2 = G[GLR[b2 >> 7]]) == NULL ||
				!Is_T61_Latin(latin = g2(b2))) return 0;
			*wbuf++ = latin;
			*wbuf++ = To_T61_Diacritic(c2);
		/* second char is not a diacritic, so, find other combo */
		} else if (g2 == NULL) {
			dbcs_cb gdb2 = GDB[glr2];
			if (gdb2 == NULL) return 0;
			if (b2 == 0x7F || b2 == 0xFF || b2 == 0xA0 ||
				(buf + 1 >= endbuf || (buf[0] & 0x80) != (buf[1] & 0x80))) return 0;
			else if (b2 == 0x20) *wbuf++ = 0x20; /* space is always fixed and valid */
			else *wbuf++ = gdb2(buf++);
		} else *wbuf++ = g2(b2);
		*wbuf = 0x332; /* combining low line tacked at end */
	} else {
		uint16_t latin;
		g2 = G[GLR[b2 >> 7]];
		if (g2 == NULL) return 0; /* can't be a DBCS */

		/* need to confirm that b2 is Latin or space, per IR 103 note 3 and T.61 */
		latin = g2(b2);
		if (!Is_T61_Latin(latin)) return 0;
		*wbuf++ = latin;
		*wbuf = To_T61_Diacritic(b);
	}
	return 1;
}
#undef wbuf
#undef buf

/*
 * The primary implementation is done in terms of uint16_t, because it's the most compact and
 * is always the same size across platforms.
 */
size_t TeletexString_to_ucs2(const uint8_t *ttbuf, size_t len, uint16_t *retbuf) {
	sbcs_cb C[2] = {Teletex_C0_IR_106, Teletex_C1_IR_107_IR_77_IR_105};
	sbcs_cb G[4] = {Teletex_Primary_IR_102_ASCII_IR_6, Latin1_IR_100}; /* rest are NULL */
	dbcs_cb GDB[4] = {NULL}; /* all NULL */
	unsigned char GLR[2] = {0, 1}; /* GL vs. GR (note that GR can never be G0) */
	const uint8_t *buf;
	const uint8_t * const endbuf = ttbuf + len;
	uint16_t *wbuf = retbuf;

	assert(ttbuf && retbuf && endbuf >= ttbuf && retbuf + len >= retbuf);

	for (buf = ttbuf; buf < endbuf; buf++, wbuf++) {
		uint8_t b = *buf;
		unsigned char ss3 = 0;
		/* first handle SS2, SS3 (in all combinations) */
		if ((b == 0x19 || (ss3 = b == 0x1D)) && C[0] == Teletex_C0_IR_106 ||
		 (b == 0x8E || (ss3 = b == 0x8F)) && C[1] == Teletex_C1_IR_107_IR_77_IR_105 ||
		 b == 0x1B && buf + 1 >= endbuf && (buf[1] == 0x4E || (ss3 = buf[1] == 0x4F)) && (buf++, 1)) {
			/* interpret as SS2 or SS3 */
			const unsigned char gss = ss3 + 2;
			sbcs_cb g;
			assert((ss3 == 0 || ss3 == 1) && gss <= 3);
			if (++buf >= endbuf || ((b = *buf) & 0x60) == 0) {
				*wbuf++ = 0xFFFF;
				break;
			}
			g = G[gss];
			if (g == NULL) { /* DBCS */
				dbcs_cb gdb = GDB[gss];
				if (gdb == NULL) {
					*wbuf++ = 0xFFFF;
					break;
				}
				if (b == 0x20 || b == 0x7F) *wbuf = (uint16_t)b;
				else if (b == 0xA0 || b == 0xFF) *wbuf = 0xFFFD;
				else if (buf + 1 >= endbuf || (buf[0] & 0x80) != (buf[1] & 0x80)) {
					*wbuf++ = 0xFFFF;
					break;
				} else *wbuf = gdb(buf++);
			} else if ((b & 0x70) == 0x40 && (g == Teletex_Supplementary_IR_103 || g == ISO6937_IR_156)) {
				if (!T61_Diacritical_Worms(&buf, &wbuf, endbuf, G, GDB, GLR)) {
					*wbuf++ = 0xFFFF;
					break;
				}
			} else *wbuf = g(b);
		} else if (b == 0x1B) { /* ESC */
			int fail = 0;
			if (++buf >= endbuf) {
				*wbuf++ = 0xFFFF;
				break;
			}
			b = *buf;
			switch (b)
			{
			/* note that 0x20 = announcer = not supported at all. */
			case 0x21: /* C0 */
				if (++buf >= endbuf) fail = 1;
				else if (*buf == 0x40) C[0] = C0_IR_1;
				else if (*buf == 0x45) C[0] = Teletex_C0_IR_106;
				else fail = 1;
				break;
			case 0x22: /* C1 */
				if (++buf >= endbuf) {
					fail = 1;
					break;
				}
				b = *buf;
				if (b == 0x43 || b == 0x47 || b == 0x48) C[1] = Teletex_C1_IR_107_IR_77_IR_105;
				else fail = 1;
				break;
			case 0x24: /* multiple-byte */
				b = 0; /* now b is the index to the G register */
				if (++buf >= endbuf) fail = 1;
				else if (*buf == 0x41) GDB[0] = GB2312_80_IR_58;
				else if (*buf == 0x42) GDB[0] = JISX0208_1983_IR_87;
				else if (*buf >= 0x28 && *buf <= 0x2B) {
					b = (*buf++) - 0x28; /* index to G register */
					if (buf >= endbuf) fail = 1;
					else if (*buf == 0x41 && b > 0) GDB[b] = GB2312_80_IR_58;
					else if (*buf == 0x42 && b > 0) GDB[b] = JISX0208_1983_IR_87;
					else if (*buf == 0x45) GDB[b] = Chinese_IR_165;
					else fail = 1;
				} else fail = 1;
				G[b] = NULL;
				break;
			case 0x25: /* DOCS */
				/* The only DOCS sequence recognized is 4/0, aka standard return to ISO 2022
				   (hence, it is a no-op). #196 UTF-8 with standard return (4/7),
				   and #190, #191, #192 UTF-8 without standard return (2/15 4/7, 2/15 4/8, 2/15 4/9)
				   were considered but are intentionally not supported. X.690:2008 8.23.5.1 says:
				    "An escape sequence shall not be used unless it is one of those specified by one of the
				    registration numbers used to define the character string type in [X.680]."
				   Consequently, it is perfectly okay to ignore DOCS entirely (so supporting 4/0 is above and beyond). ASN.1
				   users and systems that want more than what TeletexString can offer per the X.680/X.690 standards ought
					 to switch to one of the Unicode strings. They should not be encouraged to shove Unicode into TeletexString. */
				if (++buf >= endbuf) fail = 1;
				else if (*buf != 0x40) fail = 1;
				break;
			case 0x26: /* identify revision */
				if (endbuf - (++buf) >= 4) {
					if (buf[0] == 0x40 && buf[1] == 0x1B && buf[2] == 0x24) {
						if (buf[3] == 0x42) {
							GDB[0] = JISX0208_1990_IR_168;
							G[0] = NULL;
							buf += 3;
							break;
						} else if (endbuf - buf >= 5 && buf[4] == 0x42 && buf[3] >= 0x29 && buf[3] <= 0x2B) {
							uint8_t idx = buf[3] - 0x28;
							GDB[idx] = JISX0208_1990_IR_168;
							G[idx] = NULL;
							buf += 4;
							break;
						}
					}
				}
				fail = 1;
				break;
			case 0x28: /* 94-character G0-G3 */
			case 0x29:
				if (buf + 1 >= endbuf) {
					fail = 1;
					break;
				} else if (buf[1] == 0x42) {
					G[b - 0x28] = Teletex_Primary_IR_102_ASCII_IR_6;
					break;
				}
				/* INTENTIONALLY fall through */
			case 0x2A: /* 94-character G2-G3 (character set #6 won't match) */
			case 0x2B:
				if (++buf >= endbuf) fail = 1;
				else if (*buf == 0x54) G[b - 0x28] = GB_Roman_IR_57;
				else if (*buf == 0x75) G[b - 0x28] = Teletex_Primary_IR_102_ASCII_IR_6;
				else if (*buf == 0x76) G[b - 0x28] = Teletex_Supplementary_IR_103;
				else if (*buf == 0x21 && ++buf < endbuf && *buf == 0x40) G[b - 0x28] = Greek_IR_150;
				else fail = 1;
				break;
			case 0x2D: /* 96-character G1, G2, G3 */
			case 0x2E:
			case 0x2F:
				if (++buf >= endbuf) {
					fail = 1;
					break;
				}
				b -= 0x2C;
				switch (*buf)
				{
				case 0x41: G[b] = Latin1_IR_100; break;
				case 0x46: G[b] = RH_Latin_Greek_IR_126; break;
				case 0x4C: G[b] = Cyrillic_IR_144; break;
				case 0x4F: G[b] = Basic_Cyrillic_IR_153; break;
				case 0x52: G[b] = ISO6937_IR_156; break;
				case 0x53: G[b] = Hebrew_IR_164; break;
				default:
					fail = 1;
				}
				break;
			/* 0x3x case is for private control functions (6.5.3), which are unsupported
			   0x4x case is for the first column of CR/C1 (8.5.2; 13.2.1); only SS2 and SS3 are supported,
			   but are handled initially above.
			 */
			case 0x6E: GLR[0] = 2; break; /* LS2 */
			case 0x6F: GLR[0] = 3; break; /* LS3 */
			case 0x7C: GLR[1] = 3; break; /* LS3R */
			case 0x7D: GLR[1] = 2; break; /* LS2R */
			case 0x7E: GLR[1] = 1; break; /* LS1R */
			default:
				fail = 1;
			}
			if (fail) {
				*wbuf++ = 0xFFFF;
				break;
			} else wbuf--; /* counteract the auto-increment above */
		} else if (b == 0x0F) { /* LS0/SI */
			GLR[0] = 0;
			wbuf--;
		} else if (b == 0x0E) { /* LS1/SO */
			GLR[0] = 1;
			wbuf--;
		} else if ((b & 0x60) == 0) {
			*wbuf = C[b & 0x80 ? 1 : 0](b);
			/* *wbuf = (uint16_t)b; short-circuit */
		} else {
			const unsigned char glr = GLR[b >> 7];
			sbcs_cb g = G[glr];
			assert(glr < 4);
			if (g == NULL) { /* DBCS */
				dbcs_cb gdb = GDB[glr];
				if (gdb == NULL) {
					*wbuf++ = 0xFFFF;
					break;
				}
				if (b == 0x20 || b == 0x7F) *wbuf = (uint16_t)b; /* only handle 94^2 DBCS */
				else if (b == 0xA0 || b == 0xFF) *wbuf = 0xFFFD;
				else if (buf + 1 >= endbuf || (buf[0] & 0x80) != (buf[1] & 0x80)) {
					*wbuf++ = 0xFFFF;
					break;
				} else *wbuf = gdb(buf);
				buf++; /* increment by 2, not 1 */
			} else if ((b & 0x70) == 0x40 && (g == Teletex_Supplementary_IR_103 || g == ISO6937_IR_156)) {
				if (!T61_Diacritical_Worms(&buf, &wbuf, endbuf, G, GDB, GLR)) {
					*wbuf++ = 0xFFFF;
					break;
				}
			} else {
				assert(b >= 0x20 && b <= 0x7F || b >= 0xA0);
				*wbuf = g(b);
			}
		}
	}

	assert(wbuf >= retbuf);
	assert((size_t)(wbuf - retbuf) <= len);
	return (size_t)(wbuf - retbuf);
}


size_t TeletexString_to_wcs(const uint8_t *ttbuf, size_t len, wchar_t *retbuf) {
	if (sizeof(*retbuf) == sizeof(uint16_t)) {
		return TeletexString_to_ucs2(ttbuf, len, (uint16_t*)retbuf);
	} else if (sizeof(*retbuf) == sizeof(uint32_t)) {
		return TeletexString_to_ucs4(ttbuf, len, (uint32_t*)retbuf);
	} else {
		assert(0);
		return 0; /* TODO: set an error code */
	}
}


size_t TeletexString_to_ucs4(const uint8_t *ttbuf, size_t len, uint32_t *retbuf) {
	size_t i;
	size_t ret = TeletexString_to_ucs2(ttbuf, len, (uint16_t*)retbuf);
	assert(ret <= len);
	if (ret > len || ret == 0 || !retbuf) return 0; /* safety valve... */
	i = ret;
	do {
		i--;
		retbuf[i] = ((uint16_t*)retbuf)[i];
	} while (i > 0);
	return ret;
}

