/*
 * Unit Tests for ASN.1 Teletexer.
 * (C) 2011 Sean Leonard of SeanTek(R) <dev+asn1teletexer@seantek.com>.
 * For licensing information, consult the LICENSE.txt or the author.
 *
 * version 0.7, 2011-07-27
 *
 * This program executes unit tests against the ASN.1 Teletexer code.
 * It executes successful tests (where the function is expected to succeed, although
 * possibly with some U+FFFD REPLACEMENT CHARACTER values) and failure tests
 * (where the function is expected to fail with U+FFFF).
 * If successful, the program returns 0. If not successful, the program
 * returns the number of the test that failed: [1-n] for S tests; 100 + [1-n] for F tests.
 */

#include <stdio.h>
#include <memory.h>
#include <errno.h>
#include <string.h>
#include <wchar.h>

#include "../asn1teletexer.h"

const uint8_t sampletext[] = {'A', 'B', 'C', ' ', 0xA5, 0x1B, 0x24, 0x2A, 0x45,
	0x1B, 0x2F, 0x4F, 0x8E, 11 + 0x20, 28 + 0x20, 0x8E, 15 + 0x20, 8 + 0x20, 0x8E, 18 + 0xA0, 41 + 0xA0, 0x8F, 0x64 + 0x80, 0x00};

const wchar_t sampletext_u[] = {'A', 'B', 'C', ' ', 0xA5, 0x1E3F, 0x32C7, 0x91C7, 0x0444, 0x0000};

const char *RussianGreek = "\x1B\x2D\x4C\x0E\x1B\x2B\x21\x40\x1B\x7C"
	"2 aP\\^\\ QXQ[UYaZ^\\ a\\ka[U\x1B\x2A\x75\x8E, o S^R^`n R QcTciU\\ R`U\\U]X\x8E."
	"\r\n"
	"\xD3\xF4\xE7\xED \xF0\xE9\xEF \xE2\xE9\xE2\xEB\xE9\xEA\xAA \xAA\xED\xED\xEF\xE9\xE1\x8E, "
	"\xEC\xE9\xEB\xAA \xF3\xE5 \xEC\xAA\xEB\xEB\xEF\xED\xF4\xE1 \xF7\xF1\xAA\xED\xEF\x8E.";
const wchar_t *RussianGreekU = 
	L"\x0412\x00A0\x0441\x0430\x043C\x043E\x043C\x00A0\x0431\x0438\x0431\x043B\x0435\x0439\x0441"
	L"\x043A\x043E\x043C\x00A0\x0441\x043C\x044B\x0441\x043B\x0435\x002C\x00A0\x044F\x00A0\x0433"
	L"\x043E\x0432\x043E\x0440\x044E\x00A0\x0432\x00A0\x0431\x0443\x0434\x0443\x0449\x0435\x043C"
	L"\x00A0\x0432\x0440\x0435\x043C\x0435\x043D\x0438\x002E"
	L"\r\n\x03A3\x03C4\x03B7\x03BD\x00A0\x03C0\x03B9\x03BF\x00A0\x03B2\x03B9\x03B2\x03BB\x03B9\x03BA"
	L"*\x00A0*\x03BD\x03BD\x03BF\x03B9\x03B1"
	L"\x002C\x00A0\x03BC\x03B9\x03BB*\x00A0\x03C3\x03B5\x00A0\x03BC*\x03BB\x03BB\x03BF\x03BD"
	L"\x03C4\x03B1\x00A0\x03C7\x03C1*\x03BD\x03BF\x002E";

const uint8_t ChineseJapanese[] = {0x1B, 0x26, 0x40, 0x1B, 0x24, 0x2A, 0x42, 0x1B, 0x7D, /* IR #168 */
 0x1B, 0x24, 0x29, 0x41, 0x0E, /* IR #58 */
 0x3C, 0x34, 0x4A, 0x31, 0x3F, 0x6C, 0x51, 0x36, 10 + 0x20, 0x2A, 0x3D, 0x71, 0x48, 0x55, 0x52, 0x2A, 0x4E, 0x45,
 '\n', 0xBD, 0xE9, 0xB2, 0xBB, 0xA5, 0xDF, 0xA5, 0xAF, 0xA4, 0xCF, 0xB3, 0xA7, 0xA4, 0xCE, 0xC1,
 0xCF, 0xC2, 0xA4, 0xA4, 0xC7, 0xA4, 0xB9, 0xA1, 0xC1};
const uint32_t ChineseJapaneseU[] = {
	0x5373, 0x65F6, 0x5FEB, 0x8BAF, 0x2A, 0x4ECA, 0x65E5, 0x8981, 0x95FB,
	0x000A,
	/* Per JIS0208.TXT, 1-33 (A1-C1) maps to U+301C WAVE DASH, not U+FF5E FULLWIDTH TILDE. This
	   is proper Japanese punctuation. */
	0x521D, 0x97F3, 0x30DF, 0x30AF, 0x306F, 0x7686, 0x306E, 0x5275, 0x9020, 0x3067, 0x3059, 0x301C};

/* Relies on implementation detail of LS1 (0x0E) being set already */
const char UnderlineVariations[] = "\x1B\x29\x76\xCCW\xCC" "e "
	"\x8B\xCCl\xCCi\xCCk" "\xCC" "e\x8C "	"Ni\xCC\xC4n"	"\xCC" "a.";
const wchar_t UnderlineVariationsU[] = L"W\x0332" L"e\x0332 \x008Bl\x0332i\x0332k\x0332"
	L"e\x0332\x008C Nin\x0303\x0332" L"a\x0332.";
const char UnderlineChinese[] = "\x1B$(E\x1B)v\xCC" "3`\xCC~>";
const wchar_t UnderlineChineseU[] = L"\x8D64\x0332\x8345\x0332";

/* How it works:
   Tries various tests.
	 If all tests succeed, returns 0.
	 Otherwise, returns the # of the test that failed.
 */
int main(int argc, const char * argv[]) {
	size_t ret;
	wchar_t retbuf[sizeof(sampletext)];
	wchar_t retbuf2[0x80];
	size_t i = sizeof(sampletext_u)/sizeof(sampletext_u[0]);
	size_t j = wcslen(RussianGreekU);
	// size_t i, j, k;

	puts("Test S1: sample text");
	ret = TeletexString_to_wcs(sampletext, sizeof(sampletext), retbuf);

	if (ret != i || memcmp(retbuf, sampletext_u, ret)) {
		return 1;
	}

	puts("Test S2: Russian and Greek");
	if (strlen(RussianGreek) >= 0x80) return 2;
	ret = TeletexString_to_wcs((const uint8_t*)RussianGreek, strlen(RussianGreek), retbuf2);
	if (ret != j || wmemcmp(RussianGreekU, retbuf2, j)) {
		return 2;
	}

	/* 3: Chinese, Japanese */
	/* TODO: complete. this doesn't work quite yet. */
	puts("Test S3: Chinese, Japanese");
	if (1) {
		uint32_t s3buf[sizeof(ChineseJapanese)];
		wchar_t s3bufw[sizeof(ChineseJapanese)];
		ret = TeletexString_to_wcs(ChineseJapanese, sizeof(ChineseJapanese), s3bufw);
		ret = TeletexString_to_ucs4(ChineseJapanese, sizeof(ChineseJapanese), s3buf);
		if (ret != sizeof(ChineseJapaneseU)/sizeof(uint32_t) || memcmp(ChineseJapaneseU, s3buf, sizeof(uint32_t) * ret)) {
			return 3;
		}
	}

	/* 4: Enumerate TeletexString high bits */
	puts("Test S4: Enumerate TeletexString high bits");
	if (1) {
		const wchar_t *s4actual = L"\xFFFD\x00A1\x00A2\x00A3$\x00A5#\x00A7\x00A4\x2018\x201C\x00AB\x2190\x2191\x2192\x2193"
			L"\xB0\xB1\xB2\xB3\xD7\xB5\xB6\xB7\xF7\x2019\x201D\x00BB\x00BC\x00BD\x00BE\x00BF\x2015\x00B9\x00AE\x00A9"
			L"\x2122\x266A\x00AC\x00A6\xFFFD\xFFFD\xFFFD\xFFFD\x215B\x215C\x215D\x215E\x2126\x00C6\x0110\x00AA\x0126\xFFFD\x0132\x013F\x0141"
			L"\x00D8\x0152\x00BA\x00DE\x0166\x014A\x0149\x0138\x00E6\x0111\x00F0\x0127\x0131\x0133\x0140\x0142\xF8"
			L"\x0153\x00DF\x00FE\x0167\x014B\xFFFD";
		uint8_t s4bufin[16*5 + 1 + 4] = {0x1B, 0x29, 0x76, 0x0E};
		wchar_t s4bufout[sizeof(s4bufin)];
		for (i = 0xA0, j = 4; i < 0xC0; i++, j++) {
			s4bufin[j] = (uint8_t)i;
		}
		for (i = 0xD0; i < 0x100; i++, j++) {
			s4bufin[j] = (uint8_t)i;
		}

		ret = TeletexString_to_wcs(s4bufin, sizeof(s4bufin), s4bufout);

		if (ret != 0x51 || memcmp(s4bufout, s4actual, 0x51)) {
			return 4;
		}
	}

	/* 5: Try underline variations */
	puts("Test S5: Underline variations");
	if (1) {
		wchar_t s5bufout[sizeof(UnderlineVariations)] = {0};

		ret = TeletexString_to_wcs((const uint8_t*)UnderlineVariations, sizeof(UnderlineVariations), s5bufout);
		if (ret != sizeof(UnderlineVariationsU)/sizeof(UnderlineVariationsU[0]) ||
			memcmp(s5bufout, UnderlineVariationsU, sizeof(UnderlineVariationsU))) {
			return 5;
		}
	}

	/* 6: Try underline of Chinese */
	puts("Test S6: Underline of Chinese");
	if (1) {
		wchar_t s6bufout[sizeof(UnderlineChinese)] = {0};

		ret = TeletexString_to_wcs((const uint8_t*)UnderlineChinese, sizeof(UnderlineChinese), s6bufout);
		if (ret != sizeof(UnderlineChineseU)/sizeof(UnderlineChineseU[0]) ||
			memcmp(s6bufout, UnderlineChineseU, sizeof(UnderlineChineseU))) {
			return 6;
		}
	}

	/* 7: Hebrew, then interspersed Hebrew and German */
	puts("Test S7: Hebrew, then interspersed Hebrew and German");
	if (1) {
		/* Sample text says flower, then raft */
		const uint8_t s7bufin[] = {0x1B, 0x2E, 0x53, 0x1B, 0x6E, 0x1B, 0x2B, 0x75,
			0x74, 0x78, 0x67, '\t',
			0x78, 0x8F, 'F', 0x74, 0x8F, 'l', 0x71, 0x8F, 'o', 0x65, 0xDF, 0x63, 0x64, 0};
		const uint16_t s7bufUnicode[] = {0x5E4, 0x5E8, 0x5D7, '\t',
			0x5E8, 'F', 0x5E4, 'l', 0x5E1, 'o', 0x5D5, 0xDF /* eszett */, 0x5D3, 0x5D4, 0};
		uint16_t s7bufout[sizeof(s7bufin)] = {1};

		ret = TeletexString_to_ucs2(s7bufin, sizeof(s7bufin), s7bufout);
		if (ret != sizeof(s7bufUnicode)/sizeof(s7bufUnicode[0]) ||
			memcmp(s7bufUnicode, s7bufout, sizeof(s7bufUnicode))) {
			return 7;
		}
	}

	/* 8: Enumerate a whole DBCS row */
	puts("Test S8: Enumerate a whole DBCS row (ISO-IR-165)");
	if (1) {
		const wchar_t s8bufU[] = {0x4F4F, 0x6CE8, 0x795D, 0x9A7B, 0x6293, 0x722A, 0x62FD, 0x4E13, 0x7816, 0x8F6C,
			0x64B0, 0x8D5A, 0x7BC6, 0x6869, 0x5E84, 0x88C5, 0x5986, 0x649E, 0x58EE, 0x72B6, 0x690E, 0x9525, 0x8FFD, 0x8D58, 0x5760, 0x7F00,
			0x8C06, 0x51C6, 0x6349, 0x62D9, 0x5353, 0x684C, 0x7422, 0x8301, 0x914C, 0x5544, 0x7740, 0x707C, 0x6D4A, 0x5179, 0x54A8, 0x8D44,
			0x59FF, 0x6ECB, 0x6DC4, 0x5B5C, 0x7D2B, 0x4ED4, 0x7C7D, 0x6ED3, 0x5B50, 0x81EA, 0x6E0D, 0x5B57, 0x9B03, 0x68D5, 0x8E2A, 0x5B97,
			0x7EFC, 0x603B, 0x7EB5, 0x90B9, 0x8D70, 0x594F, 0x63CD, 0x79DF, 0x8DB3, 0x5352, 0x65CF, 0x7956, 0x8BC5, 0x963B, 0x7EC4, 0x94BB,
			0x7E82, 0x5634, 0x9189, 0x6700, 0x7F6A, 0x5C0A, 0x9075, 0x6628, 0x5DE6, 0x4F50, 0x67DE, 0x505A, 0x4F5C, 0x5750, 0x5EA7,
			0xFFFD, 0xFFFD, 0xFFFD, 0xFFFD, 0xFFFD};
		wchar_t s8bufout[220] = {2};
		char s8bufin[220] = "\x1B$)E\x0E";
		/* start at 4 */
		for (j = 1; j < 95; j++) {
			s8bufin[3+(j*2)] = 55 + 0x20;
			s8bufin[4+(j*2)] = 0x20 + (char)j;
		}

		ret = TeletexString_to_wcs((const uint8_t*)s8bufin, 5+94*2, s8bufout);
		if (ret != sizeof(s8bufU)/sizeof(s8bufU[0]) ||
			memcmp(s8bufU, s8bufout, sizeof(s8bufU))) {
			return 8;
		}
	}

	/* Fails: [F1] DBCS 0x80 0x00, [F2] GR without set, [F3] Announcer, [F4] invalid esc sequence,
	   [F5] DOCS UTF-8, [F6] DBCS without end char, [F7] diacritic + bad char, [F8] diacritic + end char,
		 [F9] underline + C0 */
	if (1) {
		static const char *fbufin[9] = {
			"\x1B$)E\x0E\x56\xF1",
			"\x1B\x7C\xE4",
			"\x1B \x41HUG",
			"\x1B\xF7HUG",
			"\x1B%G\xC2\xA5",
			"\x1B$(E\x56",
			"\x1B\x29\x76\x0E\xC6!",
			"\x1B\x29\x76\x0E\xC6",
			"\x1B\x29\x76\x0E\xCC\tHUG"
		};
		uint16_t fbufout[40] = {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF};

		printf("Running %u failure tests\n", (unsigned int)(sizeof(fbufin)/sizeof(fbufin[0])));
		for (i = 0; i < 9; i++) {
			ret = TeletexString_to_ucs2((const uint8_t*)fbufin[i], strlen(fbufin[i]), fbufout);
			if (ret != 1 || fbufout[0] != 0xFFFF) {
				printf("Test F%u failed\n", (unsigned int)i);
				return 101 + i;
			}
			fbufout[0] = 0xFF;
		}
	}
	
	puts("ASN.1 Teletexer tests succeeded.");
	return 0;
}

